/**
 \file com/st/p2012/mind/HTMLDocumentationHelper.java
 \brief A class providing helper methods for HTML documentation generator.


 <B>
 \n
 This file is part of the Platform 2012 program,
 a cooperation between STMicroelectronics and CEA.\n
 Redistribution of this file to outside parties is
 strictly prohibited without the written consent
 of the module owner indicated below.\n
 </B>

 \par  Module owner: michel.metzger@st.com


 \par  Copyright (C) 2009 STMicroelectronics

 \par  Authors: michel.metzger@st.com

 \par  Id: $Id$
 \par  Date: $Date$
 \par  Revision: $Rev$

 */

package com.st.p2012.mind;

import static org.objectweb.fractal.mind.NameHelper.isValid;
import static org.objectweb.fractal.mind.NameHelper.splitName;
import static org.objectweb.fractal.mind.PathHelper.fullyQualifiedNameToPath;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.mind.NameHelper;
import org.objectweb.fractal.mind.PathHelper;
import org.objectweb.fractal.mind.annotation.Annotation;
import org.objectweb.fractal.mind.annotation.AnnotationHelper;


public final class HTMLDocumentationHelper {

  public static enum SourceKind {
    COMPONENT,
    INTERFACE,
    PACKAGE
  }

  public final static String FILE_EXT = ".html";

  public final static String ADL_DOC_EXT = ".ADL" + FILE_EXT;
  public final static String ITF_DOC_EXT = ".ITF" + FILE_EXT;

  private static final String INTERFACE_ANCHOR_PREFIX = "ITF_";
  private static final String COMPONENT_ANCHOR_PREFIX = "CMP_";
  private static final String ATTRIBUTE_ANCHOR_PREFIX = "ATTR_";
  private static final String BINDING_ANCHOR_PREFIX   = "BIND_";
  private static final String IMPLEMENTATION_ANCHOR_PREFIX = "IMPL_";
  private static final String METHOD_ANCHOR_PREFIX    = "METH_";
  private static final String ENUM_ANCHOR_PREFIX      = "ENUM_";
  private static final String STRUCT_ANCHOR_PREFIX    = "STRUCT_";
  private static final String UNION_ANCHOR_PREFIX     = "UNION_";
  private static final String TYPEDEF_ANCHOR_PREFIX   = "TYPEDEF_";

  public static final String COMPONENT_SECTION_ANCHOR   = "COMPONENTS";
  public static final String CLIENT_INTERFACE_SECTION_ANCHOR   = "CLIENT_ITF";
  public static final String SERVER_INTERFACE_SECTION_ANCHOR   = "SERVER_ITF";
  public static final String ATTRIBUTE_SECTION_ANCHOR   = "ATTRIBUTES";
  public static final String BINDING_SECTION_ANCHOR = "BINDINGS";
  public static final String DATA_SECTION_ANCHOR = "DATA";
  public static final String IMPLEMENTATION_SECTION_ANCHOR = "IMPLEMENTATIONS";
  public static final String METHOD_SECTION_ANCHOR = "METHODS";
  public static final String TYPE_SECTION_ANCHOR = "TYPES";

  private HTMLDocumentationHelper() {}

  public static String getShortName(final String s) {
    final String[] sn = splitName(toValidName(s));
    return sn[sn.length - 1];
  }

  /**
   * Takes a definition name and removes template parameters
   * @param name
   * @return
   */
  public static final String toValidName(final String name) {
    final int i = name.indexOf('<');
    if (i >= 0) {
      final String defName = name.substring(0, i);
      if (!isValid(defName))
        throw new IllegalArgumentException("Name \"" + defName
            + "\" is not a valid name.");
      return defName;
    }

    if (!isValid(name))
      throw new IllegalArgumentException("Name \"" + name
          + "\" is not a valid name.");

    return name;
  }

  public static String getPathToRoot(final String name) {
    return getPackagePathToRoot(NameHelper.getPackageName(name));
  }

  /**
   * Returns the relative path to the root for a given name
   * @param name
   * @return
   */
  public static String getPackagePathToRoot(final String name) {
	if(name == null || name.length() == 0)
		return "./";
    final int depth = splitName(toValidName(name)).length;
    final StringBuilder sb = new StringBuilder();
    if (depth > 0) {
      sb.append("..");
      for (int i = 1; i < depth; i++) {
        sb.append("/..");
      }
    }
    return sb.toString();
  }

  private static String getRelativePathTo(final String sourceName, final String targetName, final SourceKind sourceKind, final String fileSuffix) {
    if(sourceKind == SourceKind.PACKAGE) {
      return getPackagePathToRoot(sourceName) + fullyQualifiedNameToPath(toValidName(targetName), fileSuffix);
    } else {
      return getPathToRoot(sourceName) + fullyQualifiedNameToPath(toValidName(targetName), fileSuffix);
    }
  }

  public static String getRelativePathToADL(final String sourceName, final String targetName, final SourceKind sourceKind) {
    return getRelativePathTo(sourceName, targetName, sourceKind, ADL_DOC_EXT);
  }

  public static String getRelativePathToITF(final String sourceName, final String targetName, final SourceKind sourceKind) {
    return getRelativePathTo(sourceName, targetName, sourceKind, ITF_DOC_EXT);
  }

  public static String getPathToADL(final String adlName) {
    return fullyQualifiedNameToPath(adlName, ADL_DOC_EXT);
  }

  public static String getPathToITF(final String itfName) {
    return fullyQualifiedNameToPath(itfName, ITF_DOC_EXT);
  }

  private static String getRelativeLinkToADLSubElem(final String sourceName,
      final String targetName, final String anchor, final SourceKind sourceKind) {
    String prefix = null;
    if(sourceName.equals(targetName) && sourceKind == SourceKind.COMPONENT) {
      prefix = "";
    } else {
      prefix = getRelativePathToADL(sourceName, targetName, sourceKind);
    }
    return prefix + (anchor.length() == 0 ? "" : "#" + anchor);
  }

  private static String getRelativeLinkToITFSubElem(final String sourceName,
      final String targetName, final String anchor, final SourceKind sourceKind) {
    String prefix = null;
    if(sourceName.equals(targetName) && sourceKind == SourceKind.INTERFACE) {
      prefix = "";
    } else {
      prefix = getRelativePathToITF(sourceName, targetName, sourceKind);
    }
    return prefix + (anchor.length() == 0 ? "" : "#" + anchor);
  }

  public static String getInterfaceAnchor(final String interfaceName) {
    return INTERFACE_ANCHOR_PREFIX + interfaceName;
  }

  public static String getSubComponentAnchor(final String componentName) {
    return COMPONENT_ANCHOR_PREFIX + componentName;
  }

  public static String getAttributeAnchor(final String attributeName) {
    return ATTRIBUTE_ANCHOR_PREFIX + attributeName;
  }

  public static String getBindingAnchor(final String fromComponent, final String fromInterface, final String toComponent, final String toInterface) {
    return String.format(BINDING_ANCHOR_PREFIX +  "%s_%s_to_%s_%s", fromComponent, fromInterface, toComponent, toInterface);
  }

  public static String getImplementationAnchor(final String path) {
    return String.format(IMPLEMENTATION_ANCHOR_PREFIX + path.replace('/', '_'));
  }

  public static String getEnumAnchor(final String enumName) {
    return ENUM_ANCHOR_PREFIX + enumName;
  }

  public static String getStructAnchor(final String structName) {
    return STRUCT_ANCHOR_PREFIX + structName;
  }

  public static String getUnionAnchor(final String unionName) {
    return UNION_ANCHOR_PREFIX + unionName;
  }

  public static String getTypedefAnchor(final String typedefName) {
    return TYPEDEF_ANCHOR_PREFIX + typedefName;
  }

  public static String getMethodAnchor(final String methodName) {
    return METHOD_ANCHOR_PREFIX + methodName;
  }

  public static String getRelativeLinkToInterface(final String sourceName, final String targetName, final String interfaceName, final SourceKind sourceKind) {
    return getRelativeLinkToADLSubElem(sourceName, targetName, getInterfaceAnchor(interfaceName), sourceKind);
  }

  public static String getRelativeLinkToSubComponent(final String sourceName, final String targetName, final String componentName, final SourceKind sourceKind) {
    return getRelativeLinkToADLSubElem(sourceName, targetName, getSubComponentAnchor(componentName), sourceKind);
  }

  public static String getRelativeLinkToAttribute(final String sourceName, final String targetName, final String attributeName, final SourceKind sourceKind) {
    return getRelativeLinkToADLSubElem(sourceName, targetName, getAttributeAnchor(attributeName), sourceKind);
  }

  public static String getRelativeLinkToInterfaceSection(
      final String sourceName, final String targetName, final SourceKind sourceKind) {
    return getRelativeLinkToADLSubElem(sourceName, targetName, CLIENT_INTERFACE_SECTION_ANCHOR + "_SUM", sourceKind);
  }

  public static String getRelativeLinkToAttributeSection(
      final String sourceName, final String targetName, final SourceKind sourceKind) {
    return getRelativeLinkToADLSubElem(sourceName, targetName, ATTRIBUTE_SECTION_ANCHOR + "_SUM", sourceKind);
  }

  public static String getRelativeLinkToSubComponentSection(
      final String sourceName, final String targetName, final SourceKind sourceKind) {
    return getRelativeLinkToADLSubElem(sourceName, targetName, COMPONENT_SECTION_ANCHOR + "_SUM", sourceKind);
  }

  public static String getRelativeLinkToStruct(final String sourceName, final String targetName, final String structName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, getStructAnchor(structName), sourceKind);
  }

  public static String getRelativeLinkToUnion(final String sourceName, final String targetName, final String unionName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, getUnionAnchor(unionName), sourceKind);
  }

  public static String getRelativeLinkToEnum(final String sourceName, final String targetName, final String enumName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, getEnumAnchor(enumName), sourceKind);
  }

  public static String getRelativeLinkToTypedef(final String sourceName, final String targetName, final String typedefName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, getTypedefAnchor(typedefName), sourceKind);
  }

  public static String getRelativeLinkToTypeSection(final String sourceName, final String targetName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, TYPE_SECTION_ANCHOR + "_SUM", sourceKind);
  }

  public static String getRelativeLinkToMethod(final String sourceName, final String targetName, final String methodName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, getMethodAnchor(methodName), sourceKind);
  }

  public static String getRelativeLinkToMethodSection(final String sourceName,final String targetName, final SourceKind sourceKind) {
    return getRelativeLinkToITFSubElem(sourceName, targetName, METHOD_SECTION_ANCHOR + "_SUM", sourceKind);
  }

  protected static String getDefinitionName(final String sourceDirectory, final String definitionFileName) throws IOException {
    final String fileName = PathHelper.removeExtension(definitionFileName);

    final String rootPath[] = sourceDirectory.split(File.separator);
    final String filePath[] = fileName.split(File.separator);

    int i;
    for(i = 0; i < rootPath.length; i++) {
      if(!filePath[i].equals(rootPath[i])) {
        throw new IllegalArgumentException("rootDirectory must be a parent of file");
      }
    }

    final StringBuilder sb = new StringBuilder();
    for(; i < filePath.length - 1; i++) {
      sb.append(filePath[i]);
      sb.append('.');
    }
    if(i < filePath.length) //necessary for empty package name
    	sb.append(filePath[i]);

    return sb.toString();
  }

  public static void addAnnotationDecoration(
      final Node container) {
    final Annotation annotations[] = AnnotationHelper.getAnnotations(container);
    if(annotations != null) {
      final List<String> annotationsList = new ArrayList<String>(annotations.length);
      for (final Annotation annotation : annotations) {
        annotationsList.add(annotation.getClass().getName());
      }
      container.astSetDecoration("annotation_list", annotationsList);
    }
  }
}
